#!/usr/bin/python
from math import pi,cos,sin,log,exp,atan
from subprocess import call
import sys, os
from Queue import Queue
#import mapnik2 as mapnik
import threading

DEG_TO_RAD = pi/180
RAD_TO_DEG = 180/pi

# Default number of rendering threads to spawn, should be roughly equal to number of CPU cores available
NUM_THREADS = 4


def minmax (a,b,c):
    a = max(a,b)
    a = min(a,c)
    return a

class GoogleProjection:
    def __init__(self,levels=18):
        self.Bc = []
        self.Cc = []
        self.zc = []
        self.Ac = []
        c = 256
        for d in range(0,levels):
            e = c/2;
            self.Bc.append(c/360.0)
            self.Cc.append(c/(2 * pi))
            self.zc.append((e,e))
            self.Ac.append(c)
            c *= 2
                
    def fromLLtoPixel(self,ll,zoom):
         d = self.zc[zoom]
         e = round(d[0] + ll[0] * self.Bc[zoom])
         f = minmax(sin(DEG_TO_RAD * ll[1]),-0.9999,0.9999)
         g = round(d[1] + 0.5*log((1+f)/(1-f))*-self.Cc[zoom])
         return (e,g)
     
    def fromPixelToLL(self,px,zoom):
         e = self.zc[zoom]
         f = (px[0] - e[0])/self.Bc[zoom]
         g = (px[1] - e[1])/-self.Cc[zoom]
         h = RAD_TO_DEG * ( 2 * atan(exp(g)) - 0.5 * pi)
         return (f,h)



class RenderThread:
    def __init__(self, tile_dir, mapfile, q, printLock, maxZoom, poly=None):
        self.tile_dir = tile_dir
        self.q = q
        #self.m = mapnik.Map(1024, 1024)
        self.printLock = printLock
        # Load style XML
        #mapnik.load_map(self.m, mapfile, True)
        # Obtain <Map> projection
        #self.prj = mapnik.Projection(self.m.srs)
        # Projects between tile pixel co-ordinates and LatLong (EPSG:4326)
        self.tileproj = GoogleProjection(maxZoom+1)
	self.poly = poly


    def render_tile(self, tile_uri, x, y, z):

        # Calculate pixel positions of bottom-left & top-right
        p0 = (x * 1024, (y + 1) * 1024)
        p1 = ((x + 1) * 1024, y * 1024)

        # Convert to LatLong (EPSG:4326)
        l0 = self.tileproj.fromPixelToLL(p0, z);
        l1 = self.tileproj.fromPixelToLL(p1, z);

	if self.poly:
		import time
		from shapely.geometry import Polygon
		p = Polygon([l0, (l0[0], l1[1]), l1, (l1[0], l0[1])])
		#sys.stdout.write("Intersect %s %s\n" % (l0,l1))
		if not self.poly.intersects(p):
			sys.stdout.write("<<< %s %s\n" % (l0,l1))
			return
	#sys.stdout.write(">>> %s %s\n" % (l0,l1))
	return

        # Convert to map projection (e.g. mercator co-ords EPSG:900913)
        c0 = self.prj.forward(mapnik.Coord(l0[0],l0[1]))
        c1 = self.prj.forward(mapnik.Coord(l1[0],l1[1]))

        # Bounding box for the tile
        if hasattr(mapnik,'mapnik_version') and mapnik.mapnik_version() >= 800:
            bbox = mapnik.Box2d(c0.x,c0.y, c1.x,c1.y)
        else:
            bbox = mapnik.Envelope(c0.x,c0.y, c1.x,c1.y)
        render_size = 1024
        self.m.resize(render_size, render_size)
        self.m.zoom_to_box(bbox)
        self.m.buffer_size = 128

        # Render image with default Agg renderer
        im = mapnik.Image(render_size, render_size)
        mapnik.render(self.m, im)
        im.save(tile_uri, 'png256')


    def loop(self):
        while True:
            #Fetch a tile from the queue and render it
            r = self.q.get()
            if (r == None):
                self.q.task_done()
                break
            else:
                (name, tile_uri, x, y, z) = r

            exists= ""
            if os.path.isfile(tile_uri):
                exists= "exists"
            else:
                self.render_tile(tile_uri, x, y, z)
            if os.path.exists(tile_uri):
                bytes=os.stat(tile_uri)[6]
                empty= ''
                if bytes == 103:
                    empty = " Empty Tile "
            else:
                empty = " Missing "
            self.printLock.acquire()
            print name, ":", z, x, y, exists, empty
            self.printLock.release()
            self.q.task_done()

def poly_parse(fp):
    from shapely.geometry import Polygon,MultiPolygon
    poly = []
    mp = []
    data = False
    for l in fp:
        l = l.strip()
        if not l: continue
        if l == 'END':
	    data = False
            if poly:
                mp.append(poly)
                poly = []
                continue
        if l.isdigit():
            data = True
            continue
        if not data: continue
        poly.append(map(lambda x: float(x.strip()), l.split()[:2]))
    if len(mp) == 1:
        return Polygon(mp[0])
    return MultiPolygon(map(Polygon, mp))

def render_tiles(bbox, mapfile, tile_dir, minZoom=1,maxZoom=18, name="unknown", num_threads=NUM_THREADS, tms_scheme=False, poly_file=None):
    print "render_tiles(",bbox, mapfile, tile_dir, minZoom,maxZoom, name,")"

    # Launch rendering threads
    queue = Queue(32)
    printLock = threading.Lock()
    renderers = {}
    poly = None
    if poly_file:
        poly = poly_parse(open(poly_file))
    for i in range(num_threads):
        renderer = RenderThread(tile_dir, mapfile, queue, printLock, maxZoom, poly=poly)
        render_thread = threading.Thread(target=renderer.loop)
        render_thread.start()
        #print "Started render thread %s" % render_thread.getName()
        renderers[i] = render_thread

    if not os.path.isdir(tile_dir):
         os.mkdir(tile_dir)

    gprj = GoogleProjection(maxZoom+1) 

    ll0 = (bbox[0],bbox[3])
    ll1 = (bbox[2],bbox[1])

    for z in range(minZoom,maxZoom + 1):
        px0 = gprj.fromLLtoPixel(ll0,z)
        px1 = gprj.fromLLtoPixel(ll1,z)

        # check if we have directories in place
        zoom = "%s" % z
        if not os.path.isdir(tile_dir + zoom):
            os.mkdir(tile_dir + zoom)
        for x in range(int(px0[0]/1024.0),int(px1[0]/1024.0)+1):
            # Validate x co-ordinate
            if (x < 0) or (x >= 2**z):
                continue
            # check if we have directories in place
            str_x = "%s" % x
            if not os.path.isdir(tile_dir + zoom + '/' + str_x):
                os.mkdir(tile_dir + zoom + '/' + str_x)
            for y in range(int(px0[1]/1024.0),int(px1[1]/1024.0)+1):
                # Validate x co-ordinate
                if (y < 0) or (y >= 2**z):
                    continue
                # flip y to match OSGEO TMS spec
                if tms_scheme:
                    str_y = "%s" % ((2**z-1) - y)
                else:
                    str_y = "%s" % y
                tile_uri = tile_dir + zoom + '/' + str_x + '/' + str_y + '.png'
                # Submit tile to be rendered into the queue
                t = (name, tile_uri, x, y, z)
                try:
                    queue.put(t)
                except KeyboardInterrupt:
                    raise SystemExit("Ctrl-c detected, exiting...")

    # Signal render threads to exit by sending empty request to queue
    for i in range(num_threads):
        queue.put(None)
    # wait for pending rendering jobs to complete
    queue.join()
    for i in range(num_threads):
        renderers[i].join()



if __name__ == "__main__":
    home = os.environ['HOME']
    try:
        mapfile = os.environ['MAPNIK_MAP_FILE']
    except KeyError:
        mapfile = home + "/OSM/mapnik/mapnik/osm.xml"
    try:
        tile_dir = os.environ['MAPNIK_TILE_DIR']
    except KeyError:
        tile_dir = "/home/site/tiles-vyatka/"

    if not tile_dir.endswith('/'):
        tile_dir = tile_dir + '/'

    #-------------------------------------------------------------------------
    #
    # Change the following for different bounding boxes and zoom levels
    #
    # Start with an overview
    # World

#    bbox = (49.9209,58.494,50.1032,58.5823)
#    render_tiles(bbox, mapfile, tile_dir, 12, 18, "Chepetsk")

#    bbox = (52.57,58.114,52.715,58.16)
#    render_tiles(bbox, mapfile, tile_dir, 10, 17, "Glazov")

#    bbox = (52.59,58.117,52.715,58.152)
#    render_tiles(bbox, mapfile, tile_dir, 18, 18, "Glazov-hirez")

#    bbox = (19.2,41,184,81)
#    render_tiles(bbox, mapfile, tile_dir, 11, 11, "Russia")

#    bbox = (49.5,58.42,50.28,58.77)
#    render_tiles(bbox, mapfile, tile_dir, 14, 14, "Chepetsk-lowzoom")

#    bbox = (49.18,58.70,49.62,59.05)
#    render_tiles(bbox, mapfile, tile_dir, 14, 15, "Vyatka")

#    bbox = (49.38,58.48,49.54,58.55)
#    render_tiles(bbox, mapfile, tile_dir, 6, 15, "Lyangasy")
    
#    bbox = (33.9554,44.3904,33.986,44.4058)
#    render_tiles(bbox, mapfile, tile_dir, 12, 18, "kazivel")

#    bbox = (49.553,58.55227,49.724,58.68)
#    render_tiles(bbox, mapfile, tile_dir, 6, 17, "Kirov1")

#    bbox = (49.56,58.49,49.70,58.58)
#    render_tiles(bbox, mapfile, tile_dir, 6, 17, "Kirov2")

#    bbox = (50.14,58.715,50.195,58.75)
#    render_tiles(bbox, mapfile, tile_dir, 6, 17, "Slobodskoy1")

#    bbox = (50.16,58.696,50.215,58.720)
#    render_tiles(bbox, mapfile, tile_dir, 6, 17, "Slobodskoy2")

#    bbox = (50.115,58.662,50.170,58.685)
#    render_tiles(bbox, mapfile, tile_dir, 6, 17, "Slobodskoy3")

#    bbox = (49.24,58.37,49.588,58.53)
#    render_tiles(bbox, mapfile, tile_dir, 13, 17, "chepetsk-raion")
########
    
    bbox = (49.9209,58.494,50.1032,58.5823)
    render_tiles(bbox, mapfile, tile_dir, 13, 18, "Chepetsk", poly_file='/tmp/vyatka.poly')
    sys.exit()


#    bbox = (46.60,55.88,54.45,60.0)
#    render_tiles(bbox, mapfile, tile_dir, 13, 16, "Vyatka")

#    bbox = (46.68,59.99,48.68,61.05)
#    render_tiles(bbox, mapfile, tile_dir, 13, 16, "NW")

    bbox = (52.59,58.117,52.715,58.152)
    render_tiles(bbox, mapfile, tile_dir, 13, 18, "Glazov-hirez")
#    sys.exit()

#    bbox = (46.60,55.88,54.45,60.0)
#    render_tiles(bbox, mapfile, tile_dir, 15, 16, "Vyatka")

#    bbox = (46.68,59.99,48.68,61.05)
#    render_tiles(bbox, mapfile, tile_dir, 15, 16, "NW")

    
#    bbox = (49.920,58.457,50.1,58.5)
#    render_tiles(bbox, mapfile, tile_dir, 16, 17, "Chepetsk-bum")
    

#    bbox = (52.57,58.114,52.715,58.16)
#    render_tiles(bbox, mapfile, tile_dir, 16, 17, "Glazov")


#    sys.exit()

    bbox = (50.14,58.715,50.195,58.75)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Slobodskoy1")

    bbox = (50.16,58.696,50.215,58.720)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Slobodskoy2")

    bbox = (50.115,58.662,50.170,58.685)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Slobodskoy3")

    bbox = (48.61,58.293,48.663,58.32)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Mirny")

    bbox = (49.553,58.55227,49.724,58.68)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Kirov1")

    bbox = (49.56,58.49,49.70,58.58)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Kirov2")

    bbox = (49.73,58.575,49.845,58.640)
    render_tiles(bbox, mapfile, tile_dir, 13, 17, "Poroshino")

